#ifndef PACULA_H
#define PACULA_H

#define MAP_WIDTH   9
#define MAP_HEIGHT 12

#include <allegro.h>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include "pacula_data.h"
using namespace std;

// Forward declarations not needed,
// but it's nice to know at a glance
// what's here  :)
struct Point;
class Cell;
class Map;

enum
{
	DIR_NONE,
	DIR_UP,
	DIR_DOWN,
	DIR_LEFT,
	DIR_RIGHT
};


struct Point
{
	int x, y;	
	
	Point(int x, int y);
	Point();	
	Point* addr();
	
	Point operator+();
	Point operator-();
	Point& operator=(const Point&);
	Point& operator+=(Point&);
	Point& operator-=(Point&);
	bool operator==(Point&);
	
	friend Point operator+(Point&, Point&);
	friend Point operator-(Point&, Point&);
};

Point operator+(Point&, Point&);
Point operator-(Point&, Point&);


class Cell
{
protected:
	bool bounds[4];
	bool pellet;
	bool powerPellet;

public:
	Cell();
	Cell(bool up,
		 bool down,
		 bool left,
		 bool right,
		 bool power);
	
	Cell(short	bounds,
		 bool	power);
	
	void draw(BITMAP*	bitmap,
			  Point*	offset,
			  Point*	cell,
			  int		cellWidth,
			  int		wallWidth,
			  int		colour);
	
	bool canMoveUp();
	bool canMoveDown();
	bool canMoveLeft();
	bool canMoveRight();
	bool hasPellet();
	bool hasPowerPellet();
	void eatPellet();
};


class Map
{
protected:
	Cell cells[MAP_WIDTH][MAP_HEIGHT];
	Point ghostHome;
	Point pacStart;
	int cellWidth;
	int wallWidth;
	Point offset;

public:
	int homingMap[MAP_WIDTH][MAP_HEIGHT];
	Map();
	Map(const char*		filename,
		int				mapIndex,
		int				cellWidth,
		int				wallWidth,
		Point*			offset);
	
	Map(FILE*	mapFile,
		int		mapIndex,
		int		cellWidth,
		int		wallWidth,
		Point*	offset);
	
	Map(FILE*	mapFile,
		int		cellWidth,
		int		wallWidth,
		Point*	offset);
	
	void buildHomingMap();
	
	void load(const char*	filename,
			  int			mapIndex);
	
	void load(FILE*	mapFile,
			 int	mapIndex);

	void load(FILE* mapFile);
	
	void draw(BITMAP*	bitmap,
			  int		colour);
	
	void drawPellets(BITMAP* bitmap, BITMAP** pellets);
	void drawPowerPellets(BITMAP* bitmap, BITMAP** pellets, int timer);
	
	Cell* getCell(Point* cell);
	
	Point* getPacStart(Point* cell);
	
	Point* getGhostHome(Point* cell);
	
	int getCellWidth();
	int getWallWidth();
	Point* getOffset(Point* offset);
};


// Creature movement controller
class Creature
{
protected:
	Map*		map;
	Point		position;			// pixel location relative to map
	Point		lastDrawPosition;
	int			direction;
	int			directionNext;
	bool		hitWall;
	int			tunnelSequence;
	BITMAP**	images;				// A set of images to use
	BITMAP*		currentImage;		// Pointer to an item in a set of images
	BITMAP*		spriteImage;		// The thing that get blitted onto the screen
	
	void decrementX(Cell* cell);
	void incrementX(Cell* cell, Point* cellPosition);
	void decrementY(Cell* cell);
	void incrementY(Cell* cell, Point* cellPosition);
	
public:
	Creature(	Map*		map,
				Point*		startCell,
			 	BITMAP**	images);
	virtual ~Creature();
	
	bool inHorizontalTransit();		// is he between cells?
	bool inVerticalTransit();
	
	// Request direction
	virtual void up();
	virtual void down();
	virtual void left();
	virtual void right();
	void setPositionCell(Point* cell);
	virtual void processSprite();
	virtual void draw(BITMAP* bitmap);
	void cleanup(BITMAP* bitmap);
	
	// Next iteration (move by one pixel)
	virtual void doFrame();
	
	void getPositionCell(Point* position);
	void getPositionPixel(Point* position);
	void getAbsolutePosition(Point* position);
	bool collide(Creature* creature);
	bool hasHitWall();
	int getDirection();
};


#endif
